/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Domain Model
   This model has the full domain model, including interface, for a lock,
but no lock.  It is used to show the relationship between the requirements
and the specification.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig UserState {}
one sig Uninterested, Waiting, InCritSec extends UserState {}

one sig Progs {  var progState: Thread -> one UserState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var lockState: Thread -> one UserState  }      -- interface

pred initialState {
   # Thread = 2                                              -- domain only
   Progs.progState = Thread -> Uninterested                  -- domain only
   Progs.runningProg = Thread                                  -- interface
   Lock.lockState = Thread -> Uninterested  }                  -- interface

/*-------------------------------------------------------------------------
ACTIONS
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {                -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock [t: Thread] { 
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.lockState' = Lock.lockState ++ t -> Waiting     -- interface change
}

pred lockReturnEnabled [t: Thread] {              -- precondition on system
   t ! in Progs.runningProg  }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   Lock.lockState' = Lock.lockState ++ t -> InCritSec   -- interface change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]  
   runningProg' = runningProg          -- no apparent interface change, but
      -- interface actually changes twice, once after each progState change
   Progs.progState' = Progs.progState ++ t -> Uninterested      -- dom chng
   Lock.lockState' = Lock.lockState ++ t -> Uninterested  -- interface chng
}

/*-------------------------------------------------------------------------
TRACES
-------------------------------------------------------------------------*/

pred doNothingEnabled { all t: Thread | 
   ! lockReturnEnabled[t] && ! unlockEnabled[t]  }
pred doNothing {
      doNothingEnabled && runningProg' = runningProg
   && progState' = progState && lockState' = lockState }

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t] || doNothing  )  }

pred anyTrace {  initialState && always delta  }
run anyTrace expect 1

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
  anyTrace => always
  (all t: Thread |    
     t.(Progs.progState) = Uninterested => 
     (  t in Progs.runningProg && t.(Lock.lockState) = Uninterested  )
  && t.(Progs.progState) = Waiting => 
        (  t ! in Progs.runningProg && t.(Lock.lockState) = Waiting  )
  && t.(Progs.progState) = InCritSec => 
     (  t in Progs.runningProg && t.(Lock.lockState) = InCritSec  )
  )  }
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread |
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1

pred faultTolerantBehavior [disj t0, t1: Thread]  {
   anyTrace
   eventually always t1.(Progs.progState) = Uninterested
   always eventually t0.(Progs.progState) = InCritSec        }
run faultTolerantBehavior expect 1

pred noSharedResource [disj t0, t1: Thread]  {
   anyTrace
   always t0.(Progs.progState) = Uninterested
   always t1.(Progs.progState) = Uninterested  }
run noSharedResource expect 1

/*-------------------------------------------------------------------------
REQUIREMENTS 
   It is not surprising that mutual exclusion and no starvation are not
satisfied, because the lock in this version has no implementation and does
not do anything.  No deadlocks is satisfied because, as the lock does
nothing, it is non-blocking.
-------------------------------------------------------------------------*/

pred domainMutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert domainMutualExclusion {  
   anyTrace => always domainMutuallyExclusive  }
check domainMutualExclusion                                 -- NOT VERIFIED

pred domainNonDeadlocking [t: Thread] {
   t.(Progs.progState) = Waiting => 
      eventually t.(Progs.progState) = InCritSec  }
assert domainNoDeadlocks {
   anyTrace => some t: Thread | always domainNonDeadlocking [t]  }
check domainNoDeadlocks                                         -- VERIFIED

assert domainNoStarvation {
   anyTrace => all t: Thread | always domainNonDeadlocking [t]  }
check domainNoStarvation                                    -- NOT VERIFIED

/*-------------------------------------------------------------------------
SPECIFICATION
   The "sufficient" assertions tell us that if the specification is
satisfied, then the requirements are satisfied.  This is the central proof
obligation of validation, although it is trivial in this case.
   There is no point to checking sufficiency of the no deadlock
specification, because in this model both specification and requirement are
already verified.
-------------------------------------------------------------------------*/

pred systemMutuallyExclusive {  # (Lock.lockState).InCritSec <= 1  }
assert systemMutualExclusion { anyTrace => always systemMutuallyExclusive }
check systemMutualExclusion                                 -- NOT VERIFIED

pred systemNonDeadlocking [t: Thread] {
   t.(Lock.lockState) = Waiting => 
      eventually t.(Lock.lockState) = InCritSec  } 
assert systemNoDeadlocks {
   anyTrace => some t: Thread | always systemNonDeadlocking [t]  }
check systemNoDeadlocks                                         -- VERIFIED

assert systemNoStarvation {
   anyTrace => all t: Thread | always systemNonDeadlocking [t]  }
check systemNoStarvation                                    -- NOT VERIFIED

assert sufficientMutexSpec { anyTrace => 
   ((always systemMutuallyExclusive) => (always domainMutuallyExclusive)) }
check sufficientMutexSpec                                       -- VERIFIED

assert sufficientNostarvSpec { anyTrace => 
   (  (all t: Thread | always systemNonDeadlocking [t]) 
   => (all t: Thread | always domainNonDeadlocking [t]) )  }
check sufficientNostarvSpec                                     -- VERIFIED
