/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 1b
   This version includes validation steps, one of which diagnoses a control
problem.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var ProgState: Thread -> one ProgState  }        -- domain

one sig Lock {  var flagRaised: set Thread  }                     -- system

pred initialState {
   # Thread = 2                                                   -- domain
   Progs.ProgState = Thread -> Uninterested                       -- domain
   no Lock.flagRaised                     }                       -- system

/*-------------------------------------------------------------------------
ACTIONS
   A manual validation check: every action must assign a new value to every
variable (framing).
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {            -- action initiated by domain
   t.(Progs.ProgState) = Uninterested  }
pred callLock [t: Thread] { 
   callLockEnabled [t]
   Progs.ProgState' = Progs.ProgState ++ t -> Waiting 
   Lock.flagRaised' = Lock.flagRaised + t           
}

pred lockReturnEnabled [t: Thread] {          -- action initiated by system
   Lock.flagRaised = t             }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   flagRaised' = flagRaised                         
   Progs.ProgState' = Progs.ProgState ++ t -> InCritSec 
}

pred unlockEnabled [t: Thread] {              -- action initiated by domain
   t.(Progs.ProgState) = InCritSec  }
pred unlock [t: Thread] {
   unlockEnabled [t]  
   Progs.ProgState' = Progs.ProgState ++ t -> Uninterested 
   Lock.flagRaised' = Lock.flagRaised - t   
}

/*-------------------------------------------------------------------------
TRACES
-------------------------------------------------------------------------*/

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t]  )  }

pred anyTrace {  initialState && always delta }
run anyTrace expect 1 

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
   Whenever the values of different state variables should be coordinated,
it is important to check that they are!  This validation step need not
respect the domain/system boundary.
-------------------------------------------------------------------------*/

assert stateInvariant {
   anyTrace => always
   (all t: Thread |    
      t.(Progs.ProgState) = Uninterested => t ! in Lock.flagRaised 
   && t.(Progs.ProgState) = Waiting => t in Lock.flagRaised 
   && t.(Progs.ProgState) = InCritSec => t in Lock.flagRaised    )  }
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
   The failure to verify sequentialControl is a problem, because we 
definitely expect actions of a thread to be sequential.  If they are not
sequential in the model, then the model could have behaviors that are
actually impossible in the real world.
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread | 
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                     -- NOT VERIFIED

/*-------------------------------------------------------------------------
REQUIREMENTS 
   Because the model is not valid, these results cannot be trusted.
-------------------------------------------------------------------------*/

pred mutuallyExclusive {  # (Progs.ProgState).InCritSec <= 1  }
assert mutualExclusion {  anyTrace => always mutuallyExclusive  }
check mutualExclusion                                           -- VERIFIED

pred nonDeadlocking {  some t: Thread | 
   (callLockEnabled [t] || lockReturnEnabled [t] || unlockEnabled [t])  }
assert noDeadlocks {  anyTrace => always nonDeadlocking  }
check noDeadlocks                                               -- VERIFIED

pred nonStarvation { all t: Thread | 
   t.(Progs.ProgState) = Waiting => 
      eventually t.(Progs.ProgState) = InCritSec  }
assert noStarvation {  anyTrace => always nonStarvation  }
check noStarvation                                              -- VERIFIED
