/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 1c
   This version fixes the control problem already found, and finds a new
problem with the domain model.
   In the last version, control was not sequential because control of the 
thread was completely implicit.  This is not adequate, so a new state 
variable is introduced.  A thread is in runningProg if and only if the
client program has control of its thread, and the lock program does not.
The variable runningProg is in the interface; it belongs to, and can be
changed by, both domain and system.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var progState: Thread -> one ProgState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var flagRaised: set Thread  }                     -- system

pred initialState {
   # Thread = 2
   Progs.runningProg = Thread                     
   Progs.progState = Thread -> Uninterested
   no Lock.flagRaised                     }

/*-------------------------------------------------------------------------
ACTIONS
   Now the actions use the new control variable.  The transfer of control 
of a thread between client program and lock program is explicitly recorded
in the interface variable.
   The comments record our manual validation that domain and system are
kept separate.  
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {                -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock [t: Thread] { 
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.flagRaised' = Lock.flagRaised + t                  -- system change
}

pred lockReturnEnabled [t: Thread] {              -- precondition on system
   Lock.flagRaised = t && t ! in Progs.runningProg }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   flagRaised' = flagRaised                             -- no system change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]  
   runningProg' = runningProg              -- no apparent interface change,
      -- but interface actually changes twice, once after each state change
   Progs.progState' = Progs.progState ++ t -> Uninterested    -- dom change
   Lock.flagRaised' = Lock.flagRaised - t                  -- system change
}

/*-------------------------------------------------------------------------
TRACES
-------------------------------------------------------------------------*/

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t]  )  }

pred anyTrace {  initialState && always delta }
run anyTrace expect 1 

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
   anyTrace => always
   (all t: Thread |    
      t.(Progs.progState) = Uninterested => 
         (  t in Progs.runningProg && t ! in Lock.flagRaised  )
   && t.(Progs.progState) = Waiting => 
         (  t ! in Progs.runningProg && t in Lock.flagRaised  )
   && t.(Progs.progState) = InCritSec => 
         (  t in Progs.runningProg && t in Lock.flagRaised    )  )  }
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
   Now control is sequential.
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread | 
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
   A very basic validation concept: check that the domain has the expected
freedom.  The inconsistency of thereIsContention reveals a validation bug.
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1                              -- INCONSISTENT

/*-------------------------------------------------------------------------
REQUIREMENTS 
   Because the model is not valid, these results cannot be trusted.
-------------------------------------------------------------------------*/

pred mutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert mutualExclusion {  anyTrace => always mutuallyExclusive  }
check mutualExclusion                                           -- VERIFIED

pred nonDeadlocking {  some t: Thread | 
   (callLockEnabled [t] || lockReturnEnabled [t] || unlockEnabled [t])  }
assert noDeadlocks {  anyTrace => always nonDeadlocking  }
check noDeadlocks                                               -- VERIFIED

pred nonStarvation { all t: Thread | 
   t.(Progs.progState) = Waiting => 
      eventually t.(Progs.progState) = InCritSec  }
assert noStarvation {  anyTrace => always nonStarvation  }
check noStarvation                                              -- VERIFIED
