/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 1d
   This version fixes the validity problem with the domain model found in 
Version 1c, and discusses a new one.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var progState: Thread -> one ProgState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var flagRaised: set Thread  }                     -- system

pred initialState {
   # Thread = 2
   Progs.runningProg = Thread
   Progs.progState = Thread -> Uninterested
   no Lock.flagRaised                     }

/*-------------------------------------------------------------------------
ACTIONS
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {                -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock [t: Thread] {
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.flagRaised' = Lock.flagRaised + t                  -- system change
}

pred lockReturnEnabled [t: Thread] {              -- precondition on system
   Lock.flagRaised = t && t ! in Progs.runningProg }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   flagRaised' = flagRaised                             -- no system change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]
   runningProg' = runningProg              -- no apparent interface change,
      -- but interface actually changes twice, once after each state change
   Progs.progState' = Progs.progState ++ t -> Uninterested    -- dom change
   Lock.flagRaised' = Lock.flagRaised - t                  -- system change
}

/*-------------------------------------------------------------------------
TRACES
   Previous validation revealed that the model allows no contention for the
shared resource, which makes it pretty useless!
   The problem is that there were no stuttering steps, so, combined with
the restriction to infinite traces, anyTrace actually guaranteed the
absence of deadlocks.  The addition of a stuttering step solves this 
problem.
-------------------------------------------------------------------------*/

pred doNothingEnabled { all t: Thread | 
   ! callLockEnabled[t] && ! lockReturnEnabled[t] && ! unlockEnabled[t]  }
pred doNothing {  
   doNothingEnabled && runningProg' = runningProg && 
   progState' = progState && flagRaised' = flagRaised  }

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t] || doNothing  )  }

pred anyTrace {  initialState && always delta  }
run anyTrace expect 1 

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
   anyTrace => always
   (all t: Thread |
      t.(Progs.progState) = Uninterested =>
         (  t in Progs.runningProg && t ! in Lock.flagRaised  )
   && t.(Progs.progState) = Waiting =>
         (  t ! in Progs.runningProg && t in Lock.flagRaised  )
   && t.(Progs.progState) = InCritSec =>
         (  t in Progs.runningProg && t in Lock.flagRaised    )  )  }
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread | 
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1

/*-------------------------------------------------------------------------
REQUIREMENTS 
   Finally, the truth is that Lock 1 allows deadlocks and starvation.
   Our notions of validity depend, not just on realism, but also on proper
separation of domain and system.  From this perspective, the requirements
are not completely satisfactory:
   No deadlocks is partly defined in terms of the lock implementation (the
lock variable flagRaised, references in lockReturnEnabled), which 
contradicts the rule that requirements should constrain domain phenomena 
only (including the domain/system interface, which is shared between them).
-------------------------------------------------------------------------*/

pred mutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert mutualExclusion {  anyTrace => always mutuallyExclusive  }
check mutualExclusion                                           -- VERIFIED

pred nonDeadlocking {  some t: Thread |
   (callLockEnabled [t] || lockReturnEnabled [t] || unlockEnabled [t])  }
assert noDeadlocks {  anyTrace => always nonDeadlocking  }
check noDeadlocks                                           -- NOT VERIFIED

pred nonStarvation { all t: Thread |
   t.(Progs.progState) = Waiting =>
      eventually t.(Progs.progState) = InCritSec  }
assert noStarvation {  anyTrace => always nonStarvation  }
check noStarvation                                          -- NOT VERIFIED
