/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 2a
   This version models Herlihy's Lock 2.  It also rewrites the requirements
so that they do not depend on the implementation.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var progState: Thread -> one ProgState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var polite: lone Thread  }                        -- system

pred initialState {
   # Thread = 2
   Progs.runningProg = Thread
   Progs.progState = Thread -> Uninterested
   no Lock.polite                         }

/*-------------------------------------------------------------------------
ACTIONS
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {                -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock [t: Thread] {
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.polite' = t                                        -- system change
}

pred lockReturnEnabled [t: Thread] {              -- precondition on system
   Lock.polite != t && t ! in Progs.runningProg }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   polite' = polite                                     -- no system change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]
   runningProg' = runningProg              -- no apparent interface change,
      -- but interface actually changes twice, once after each state change
   Progs.progState' = Progs.progState ++ t -> Uninterested    -- dom change
   polite' = polite                                     -- no system change
}

/*-------------------------------------------------------------------------
TRACES
-------------------------------------------------------------------------*/

pred doNothingEnabled { all t: Thread | 
   ! callLockEnabled[t] && ! lockReturnEnabled[t] && ! unlockEnabled[t]  }
pred doNothing {
      doNothingEnabled && runningProg' = runningProg
   && progState' = progState && polite' = polite   }

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t] || doNothing  )  }

pred anyTrace {  initialState && always delta  }
run anyTrace expect 1

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
   anyTrace => always
   (all t: Thread |
      t.(Progs.progState) = Uninterested =>
         (  t in Progs.runningProg && Lock.polite != t  )
   && t.(Progs.progState) = Waiting => t ! in Progs.runningProg 
   && t.(Progs.progState) = InCritSec =>
         (  t in Progs.runningProg && Lock.polite ! = t)  )  }
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread | 
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
   With Lock 2 we introduce a new domain scenario, in which one thread 
dies for some reason and never again requests the lock.  For fault 
tolerance, the other thread should be able to operate normally.
   The predicate encoding this scenario cannot be instantiated, which
pinpoints what is wrong with Lock 2: a thread cannot enter its critical
section unless the other thread is requesting its critical section around
the same time!
   There is another new domain scenario, in which there is no shared
resources and the client programs never need to use the lock.  This is
inconsistent also, which does not seem to make sense.
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1 

pred faultTolerantBehavior [disj t0, t1: Thread]  {
   anyTrace
   eventually always t1.(Progs.progState) = Uninterested
   always eventually t0.(Progs.progState) = InCritSec  }
run faultTolerantBehavior expect 1                          -- INCONSISTENT

pred noSharedResource [disj t0, t1: Thread]  {
   anyTrace
   always t0.(Progs.progState) = Uninterested
   always t1.(Progs.progState) = Uninterested  }
run noSharedResource expect 1                               -- INCONSISTENT

/*-------------------------------------------------------------------------
REQUIREMENTS 
   Our notions of validity depend, not just on realism, but also on proper
separation of domain and system.  From this perspective, the requirements
on Version 1 were not completely satisfactory:
   No deadlocks was partly defined in terms of the lock implementation (the
lock variable flagRaised, referenced in lockReturnEnabled), which 
contradicted the rule that requirements should constrain domain phenomena 
only (including the domain/system interface, which is shared between them).
   Here the requirements are written in a different way, which does not
depend on the implementation.
   Lock 2 satisfies the requirements, but with such a huge flaw in its
validity, that does not matter much.
-------------------------------------------------------------------------*/

pred mutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert mutualExclusion {  anyTrace => always mutuallyExclusive  }
check mutualExclusion                                           -- VERIFIED

pred nonDeadlocking [t: Thread] {
   t.(Progs.progState) = Waiting =>
      eventually t.(Progs.progState) = InCritSec  }
assert noDeadlocks {  
   anyTrace => some t: Thread | always nonDeadlocking [t]  }
check noDeadlocks                                               -- VERIFIED

assert noStarvation {  
   anyTrace => all t: Thread | always nonDeadlocking [t]  }
check noStarvation                                              -- VERIFIED
