/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 2b
   This version discusses and improves the noSharedResource domain
behavior.
   This version also comments on the lack of a true specification.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var progState: Thread -> one ProgState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var polite: lone Thread  }                        -- system

pred initialState {
   # Thread = 2
   Progs.runningProg = Thread
   Progs.progState = Thread -> Uninterested
   no Lock.polite                         }

/*-------------------------------------------------------------------------
ACTIONS
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {                -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock [t: Thread] {
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.polite' = t                                        -- system change
}

pred lockReturnEnabled [t: Thread] {              -- precondition on system
   Lock.polite != t && t ! in Progs.runningProg }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   polite' = polite                                     -- no system change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]
   runningProg' = runningProg              -- no apparent interface change,
      -- but interface actually changes twice, once after each state change
   Progs.progState' = Progs.progState ++ t -> Uninterested    -- dom change
   polite' = polite                                     -- no system change
}

/*-------------------------------------------------------------------------
TRACES
   Finally, we have stuttering correct in the domain model!  There cannot
be gratuitous stuttering steps when the locking mechanism has something to
do, because if there were, it would be impossible to prove progress
properties.  However, it is necessary to allow stuttering when the locking
mechanism has nothing to do, so that legal (infinite) traces are allowed
when there is nothing to do.
-------------------------------------------------------------------------*/

pred doNothingEnabled { all t: Thread | 
   ! lockReturnEnabled[t] && ! unlockEnabled[t]  }
pred doNothing {
      doNothingEnabled && runningProg' = runningProg
   && progState' = progState && polite' = polite   }

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t] || doNothing  )  }

pred anyTrace {  initialState && always delta  }
run anyTrace expect 1

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
   anyTrace => always
   (all t: Thread |
      t.(Progs.progState) = Uninterested =>
         (  t in Progs.runningProg && Lock.polite != t  )
   && t.(Progs.progState) = Waiting => t ! in Progs.runningProg 
   && t.(Progs.progState) = InCritSec =>
         (  t in Progs.runningProg && Lock.polite ! = t)  )  }
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread | 
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
   Now all these validation results make sense.
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1 

pred faultTolerantBehavior [disj t0, t1: Thread]  {
   anyTrace
   eventually always t1.(Progs.progState) = Uninterested
   always eventually t0.(Progs.progState) = InCritSec  }
run faultTolerantBehavior expect 1                          -- INCONSISTENT

pred noSharedResource [disj t0, t1: Thread]  {
   anyTrace
   always t0.(Progs.progState) = Uninterested
   always t1.(Progs.progState) = Uninterested  }
run noSharedResource expect 1 

/*-------------------------------------------------------------------------
REQUIREMENTS 
   As before, the model is not valid, so it does not matter what properties
can and cannot be verified.
   Exercise: Note that noStarvation fails now, while it was verified in
Lock 2a.  Can you explain why?
   Our notions of validity depend, not just on realism, but also on proper
separation of domain and system.  From this perspective, there is a
remaining problem:  Where is the specification?  Ideally there should be a
specification whose phenomena are observable by the system, yet do not
constrain the lock implementation--because we don't know what it is yet.
-------------------------------------------------------------------------*/

pred mutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert mutualExclusion {  anyTrace => always mutuallyExclusive  }
check mutualExclusion                                           -- VERIFIED

pred nonDeadlocking [t: Thread] {
   t.(Progs.progState) = Waiting =>
      eventually t.(Progs.progState) = InCritSec  }
assert noDeadlocks {
   anyTrace => some t: Thread | always nonDeadlocking [t]  }
check noDeadlocks                                               -- VERIFIED

assert noStarvation {
   anyTrace => all t: Thread | always nonDeadlocking [t]  }
check noStarvation                                          -- NOT VERIFIED
