/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 3a (Peterson Lock)
   In a very direct sense, this model is the sum of Lock 1 and Lock 2.
   In response to comments about Lock 2, this version supports a real
specification.  The interface variable lockState tracks progState.  It is
visible to the system, and updated by the system.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, InCritSec extends ProgState {}

one sig Progs {  var progState: Thread -> one ProgState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var lockState: Thread -> one ProgState,        -- interface
                var flagRaised: set Thread,                  -- system only
                var polite: lone Thread               }      -- system only

pred initialState {
   # Thread = 2                                              -- domain only
   Progs.progState = Thread -> Uninterested                  -- domain only
   Lock.lockState = Thread -> Uninterested                     -- interface
   Progs.runningProg = Thread                                  -- interface
   no Lock.flagRaised && no Lock.polite   }                  -- system only

/*-------------------------------------------------------------------------
ACTIONS
   Lock 3 maintains the states of both Lock 1 and Lock 2.  Crucially, the
enabling condition on lockReturn is the OR of the previous enabling
conditions, not the AND.  This is how it gets the benefits of both previous
designs, without their disadvantages.
   With this version comes a new validity concern, which is atomicity.
In an implementation, flagRaised and polite would be represented by three
variables: a boolean for each thread indicating whether its flag is raised,
and an optional thread value for polite.  callLock changes two of these
variables.  The precondition of lockReturn reads two of these variables
(assuming it does not read the flagRaised boolean for its own thread).  And
all three variables can be read and written concurrently by actions of the
other thread.  It does not seem that these actions are guaranteed to be
atomic in the implementation.
-------------------------------------------------------------------------*/

pred callLockEnabled [t: Thread] {                -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock [t: Thread] { 
   callLockEnabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.lockState' = Lock.lockState ++ t -> Waiting     -- interface change
   Lock.flagRaised' = Lock.flagRaised + t                  -- system change
   Lock.polite' = t                                        -- system change
}

pred lockReturnEnabled [t: Thread] {              -- precondition on system
      (  Lock.flagRaised = t || Lock.polite ! = t )
   && t ! in Progs.runningProg                    }
pred lockReturn [t: Thread] {
   lockReturnEnabled [t]
   flagRaised' = flagRaised                             -- no system change
   polite' = polite                                     -- no system change
   Lock.lockState' = Lock.lockState ++ t -> InCritSec   -- interface change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]  
   runningProg' = runningProg          -- no apparent interface change, but
      -- interface actually changes twice, once after each progState change
   Progs.progState' = Progs.progState ++ t -> Uninterested      -- dom chng
   Lock.lockState' = Lock.lockState ++ t -> Uninterested --interface change
   Lock.flagRaised' = Lock.flagRaised - t                  -- system change
   polite' = polite                                     -- no system change
}

/*-------------------------------------------------------------------------
TRACES
-------------------------------------------------------------------------*/

pred doNothingEnabled { all t: Thread | 
   ! lockReturnEnabled[t] && ! unlockEnabled[t]  }
pred doNothing {
      doNothingEnabled && runningProg' = runningProg
   && progState' = progState && lockState' = lockState 
   && flagRaised' = flagRaised && polite' = polite      }

pred delta {  some t: Thread |
   (  callLock [t] || lockReturn [t] || unlock [t] || doNothing  )  }

pred anyTrace {  initialState && always delta  }
run anyTrace expect 1

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
  anyTrace => always
  (all t: Thread |    
     t.(Progs.progState) = Uninterested => 
     (  t in Progs.runningProg && t.(Lock.lockState) = Uninterested 
     && t ! in Lock.flagRaised && Lock.polite != t                )
  && t.(Progs.progState) = Waiting => 
        (  t ! in Progs.runningProg && t.(Lock.lockState) = Waiting
        && t in Lock.flagRaised                                   )
  && t.(Progs.progState) = InCritSec => 
     (  t in Progs.runningProg && t.(Lock.lockState) = InCritSec
     && t in Lock.flagRaised && Lock.polite != t               )
  )                                                               
}
check stateInvariant                                            -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread |
      ! (callLockEnabled [t] && lockReturnEnabled [t])
   && ! (callLockEnabled [t] && unlockEnabled [t])
   && ! (lockReturnEnabled [t] && unlockEnabled [t]) )  }
check sequentialControl                                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1 

pred faultTolerantBehavior [disj t0, t1: Thread]  {
   anyTrace
   eventually always t1.(Progs.progState) = Uninterested
   always eventually t0.(Progs.progState) = InCritSec        }
run faultTolerantBehavior expect 1 

pred noSharedResource [disj t0, t1: Thread]  {
   anyTrace
   always t0.(Progs.progState) = Uninterested
   always t1.(Progs.progState) = Uninterested  }
run noSharedResource expect 1

/*-------------------------------------------------------------------------
REQUIREMENTS 
   Now that faultTolerantBehavior can be instantiated, ask yourself this:
   Exercise: From the perspective of the domain, why is an instance of
faultTolerantBehavior different from an instance of a trace (not allowed
by this model) that satisfies domainNoDeadlocks but is a counterexample to
domainNoStarvation?
-------------------------------------------------------------------------*/

pred domainMutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert domainMutualExclusion {  
   anyTrace => always domainMutuallyExclusive  }
check domainMutualExclusion                                     -- VERIFIED

pred domainNonDeadlocking [t: Thread] {
   t.(Progs.progState) = Waiting => 
      eventually t.(Progs.progState) = InCritSec  }
assert domainNoDeadlocks {
   anyTrace => some t: Thread | always domainNonDeadlocking [t]  }
check domainNoDeadlocks                                         -- VERIFIED

assert domainNoStarvation {
   anyTrace => all t: Thread | always domainNonDeadlocking [t]  }
check domainNoStarvation                                        -- VERIFIED

/*-------------------------------------------------------------------------
SPECIFICATION
-------------------------------------------------------------------------*/

pred systemMutuallyExclusive {  # (Lock.lockState).InCritSec <= 1  }
assert systemMutualExclusion { anyTrace => always systemMutuallyExclusive }
check systemMutualExclusion                                     -- VERIFIED

pred systemNonDeadlocking [t: Thread] {
   t.(Lock.lockState) = Waiting =>
      eventually t.(Lock.lockState) = InCritSec  } 
assert systemNoDeadlocks {
   anyTrace => some t: Thread | always systemNonDeadlocking [t]  }
check systemNoDeadlocks                                         -- VERIFIED

assert systemNoStarvation {
   anyTrace => all t: Thread | always systemNonDeadlocking [t]  }
check systemNoStarvation                                        -- VERIFIED
