/*-------------------------------------------------------------------------
AN EXPERIMENT IN VALIDATION, Lock Version 3b (Peterson Lock)
   This version improves on 3a by having atomic actions.  The new lockState
variable becomes more useful, as it can store the additional state needed
to coordinate more actions.
   For extra validation, we proved the theorems with longer traces (and
found no new problems).
   Was all the fuss about atomic actions pointless?  Sure, we've all been
told that the Peterson lock works.  But if this were a new lock design, and
nobody knew its properties for sure, atomic actions would be very
important.
-------------------------------------------------------------------------*/

sig Thread {}

abstract sig ProgState {}
one sig Uninterested, Waiting, Waiting2, InCritSec extends ProgState {}
                                                        -- note extra value

one sig Progs {  var progState: Thread -> one ProgState,          -- domain
                 var runningProg: set Thread           }       -- interface

one sig Lock {  var lockState: Thread -> one ProgState,        -- interface
                var flagRaised: set Thread,                  -- system only
                var polite: lone Thread               }      -- system only

pred initialState {
   # Thread = 2                                              -- domain only
   Progs.progState = Thread -> Uninterested                  -- domain only
   Lock.lockState = Thread -> Uninterested                     -- interface
   Progs.runningProg = Thread                                  -- interface
   no Lock.flagRaised && no Lock.polite   }                  -- system only

/*-------------------------------------------------------------------------
ACTIONS
   In this version there are five actions instead of three, but they can
be manually validated as atomic.
-------------------------------------------------------------------------*/

pred callLock1Enabled [t: Thread] {               -- precondition on domain
   t.(Progs.progState) = Uninterested && t in Progs.runningProg  }
pred callLock1 [t: Thread] { 
   callLock1Enabled [t]
   Progs.progState' = Progs.progState ++ t -> Waiting      -- domain change
   Progs.runningProg' = Progs.runningProg - t           -- interface change
   Lock.lockState' = Lock.lockState ++ t -> Waiting     -- interface change
   Lock.flagRaised' = Lock.flagRaised + t                  -- system change
   polite' = polite                                     -- no system change
}

pred callLock2Enabled [t: Thread] {               -- precondition on system
   t.(Lock.lockState) = Waiting && t ! in Progs.runningProg  }
pred callLock2 [t: Thread] { 
   callLock2Enabled [t]
   progState' = progState                               -- no domain change
   runningProg' = runningProg                        -- no interface change
   Lock.lockState' = Lock.lockState ++ t -> Waiting2    -- interface change
   flagRaised' = flagRaised                             -- no system change
   Lock.polite' = t                                        -- system change
}

pred lockReturn1Enabled [t: Thread] {             -- precondition on system
      t.(Lock.lockState) = Waiting2 && t ! in Progs.runningProg 
   && Lock.flagRaised = t                                     }
pred lockReturn1 [t: Thread] {
   lockReturn1Enabled [t]
   flagRaised' = flagRaised                             -- no system change
   polite' = polite                                     -- no system change
   Lock.lockState' = Lock.lockState ++ t -> InCritSec   -- interface change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred lockReturn2Enabled [t: Thread] {             -- precondition on system
      t.(Lock.lockState) = Waiting2 && t ! in Progs.runningProg 
   && Lock.polite ! = t                                       }
pred lockReturn2 [t: Thread] {
   lockReturn2Enabled [t]
   flagRaised' = flagRaised                             -- no system change
   polite' = polite                                     -- no system change
   Lock.lockState' = Lock.lockState ++ t -> InCritSec   -- interface change
   Progs.runningProg' = Progs.runningProg + t           -- interface change
   Progs.progState' = Progs.progState ++ t -> InCritSec    -- domain change
}

pred unlockEnabled [t: Thread] {                  -- precondition on domain
   t.(Progs.progState) = InCritSec && t in Progs.runningProg }
pred unlock [t: Thread] {
   unlockEnabled [t]  
   Progs.progState' = Progs.progState ++ t -> Uninterested      -- dom chng
   runningProg' = runningProg          -- no apparent interface change, but
      -- interface actually changes twice, once after each progState change
   Lock.lockState' = Lock.lockState ++ t -> Uninterested --interface change
   Lock.flagRaised' = Lock.flagRaised - t                  -- system change
   polite' = polite                                     -- no system change
}

/*-------------------------------------------------------------------------
TRACES
   By default, Alloy 6 explores all possible lasso traces up to 10 states 
or actions.  What if the results we get are wrong because there are
counterexamples longer than this?
   Our basic idea is that now, with finer-grained actions, each thread
goes through a complete cycle in 4 actions.  Both threads cycle completely
in 8 actions, which was formally 6.  The difference between 6 and 10 gives
some headroom, but there is less in 8 vs. 10.  This might be a weakness in 
the validation, so we are analyzing traces up to 16 actions (steps).  
-------------------------------------------------------------------------*/

pred doNothingEnabled { all t: Thread |
   ! callLock2Enabled [t] &&
   !lockReturn1Enabled[t] && !lockReturn2Enabled[t] && !unlockEnabled[t]  }
pred doNothing {
      doNothingEnabled && runningProg' = runningProg
   && progState' = progState && lockState' = lockState 
   && flagRaised' = flagRaised && polite' = polite   }

pred delta {  some t: Thread |
   (  callLock1 [t] || callLock2 [t] || lockReturn1 [t]
   || lockReturn2 [t] || unlock [t] || doNothing      )  }

pred anyTrace {  initialState && always delta  }
run anyTrace expect 1

/*-------------------------------------------------------------------------
VALIDATION: INVARIANT CHECKS
-------------------------------------------------------------------------*/

assert stateInvariant {
  anyTrace => always
  (all t: Thread |    
     t.(Progs.progState) = Uninterested => 
     (  t in Progs.runningProg && t.(Lock.lockState) = Uninterested 
     && t ! in Lock.flagRaised && Lock.polite != t                )
  && t.(Progs.progState) = Waiting => 
        (  t ! in Progs.runningProg && t in Lock.flagRaised 
        && t.(Lock.lockState) in ( Waiting + Waiting2 )   )
  && t.(Progs.progState) = InCritSec => 
     (  t in Progs.runningProg && t.(Lock.lockState) = InCritSec
     && t in Lock.flagRaised && Lock.polite != t               )
  )  }
check stateInvariant for 2 but 16 steps                         -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: CONTROL CHECKS
-------------------------------------------------------------------------*/

assert sequentialControl {
   anyTrace => always
   (all t: Thread |
      ! (callLock1Enabled [t] && callLock2Enabled [t])
   && ! (callLock1Enabled [t] && lockReturn1Enabled [t])
   && ! (callLock1Enabled [t] && lockReturn2Enabled [t])
   && ! (callLock1Enabled [t] && unlockEnabled [t])
   && ! (callLock2Enabled [t] && lockReturn1Enabled [t])
   && ! (callLock2Enabled [t] && lockReturn2Enabled [t])
   && ! (callLock2Enabled [t] && unlockEnabled [t])
   && ! (lockReturn1Enabled [t] && lockReturn2Enabled [t])
   && ! (lockReturn1Enabled [t] && unlockEnabled [t])
   && ! (lockReturn2Enabled [t] && unlockEnabled [t])    )   }
check sequentialControl for 2 but 16 steps                      -- VERIFIED

/*-------------------------------------------------------------------------
VALIDATION: DOMAIN BEHAVIOR
-------------------------------------------------------------------------*/

pred bothProgsRequestLock [disj t0, t1: Thread]  {
   anyTrace
   eventually t0.(Progs.progState) = Waiting
   eventually t1.(Progs.progState) = Waiting  }
run bothProgsRequestLock expect 1

pred thereIsContention [disj t0, t1: Thread]  {
   anyTrace
   eventually Thread.(Progs.progState) = Waiting  }
run thereIsContention expect 1

pred faultTolerantBehavior [disj t0, t1: Thread]  {
   anyTrace
   eventually always t1.(Progs.progState) = Uninterested
   always eventually t0.(Progs.progState) = InCritSec        }
run faultTolerantBehavior expect 1

pred noSharedResource [disj t0, t1: Thread]  {
   anyTrace
   always t0.(Progs.progState) = Uninterested
   always t1.(Progs.progState) = Uninterested  }
run noSharedResource expect 1

/*-------------------------------------------------------------------------
REQUIREMENTS 
-------------------------------------------------------------------------*/

pred domainMutuallyExclusive {  # (Progs.progState).InCritSec <= 1  }
assert domainMutualExclusion {  
   anyTrace => always domainMutuallyExclusive  }
check domainMutualExclusion for 2 but 16 steps                  -- VERIFIED

pred domainNonDeadlocking [t: Thread] {
   t.(Progs.progState) = Waiting => 
      eventually t.(Progs.progState) = InCritSec  }
assert domainNoDeadlocks {
   anyTrace => some t: Thread | always domainNonDeadlocking [t]  }
check domainNoDeadlocks for 2 but 16 steps                      -- VERIFIED

assert domainNoStarvation {
   anyTrace => all t: Thread | always domainNonDeadlocking [t]  }
check domainNoStarvation for 2 but 16 steps                     -- VERIFIED

/*-------------------------------------------------------------------------
SPECIFICATION
-------------------------------------------------------------------------*/

pred systemMutuallyExclusive {  # (Lock.lockState).InCritSec <= 1  }
assert systemMutualExclusion { anyTrace => always systemMutuallyExclusive }
check systemMutualExclusion for 2 but 16 steps                  -- VERIFIED

pred systemNonDeadlocking [t: Thread] {
   t.(Lock.lockState) = Waiting2 => 
      eventually t.(Lock.lockState) = InCritSec  }
assert systemNoDeadlocks {
   anyTrace => some t: Thread | always systemNonDeadlocking [t]  }
check systemNoDeadlocks for 2 but 16 steps                      -- VERIFIED

assert systemNoStarvation {
   anyTrace => all t: Thread | always systemNonDeadlocking [t]  }
check systemNoStarvation for 2 but 16 steps                     -- VERIFIED
